/*
 * $Id: rip_proto.c,v 1.7 1998/09/04 20:43:54 masaki Exp $
 * routines shared by RIP-2 and RIPng
 */

#include <stdarg.h>
#include <mrt.h>
#include <api6.h>
#include <array.h>
#include <rip.h>

rip_attr_t *
rip_new_attr (rip_t * rip, int metric)
{
    rip_attr_t *attr;

    attr = New (rip_attr_t);
    attr->type = rip->proto;
    attr->metric = metric;
    attr->ref_count = 1;
    attr->pref = RIP_PREF;
    time (&attr->time);
    pthread_mutex_init (&attr->mutex_lock, NULL);

    return (attr);
}


rip_attr_t *
rip_ref_attr (rip_attr_t * attr)
{
    pthread_mutex_lock (&attr->mutex_lock);
    assert (attr->ref_count > 0);
    attr->ref_count++;
    pthread_mutex_unlock (&attr->mutex_lock);
    return (attr);
}


void
rip_deref_attr (rip_attr_t * attr)
{
    pthread_mutex_lock (&attr->mutex_lock);
    assert (attr->ref_count > 0);
    if (--attr->ref_count <= 0) {
	Deref_Prefix (attr->nexthop);
	pthread_mutex_destroy (&attr->mutex_lock);
	Delete (attr);
	return;
    }
    pthread_mutex_unlock (&attr->mutex_lock);
}


/* 
 * create/allocate new rip_route_t structure and insert
 * it in the prefix route hash table
 */
rip_route_t *
rip_new_route (rip_t * rip, prefix_t * prefix, rip_attr_t * attr)
{
    rip_route_t *route;

    route = New (rip_route_t);
    route->prefix = Ref_Prefix (prefix);
    route->attr = rip_ref_attr (attr);
    route->time = time (NULL);
    route->delete = 0;
    route->flags |= RT_RIP_CHANGE;
    route->imported = NULL;

    HASH_Insert (rip->hash, route);
    rip->num_routes++;
    return (route);
}


static rip_route_t *
rip_new_imported_route (rip_t * rip, prefix_t * prefix, rip_attr_t * attr)
{
    rip_route_t *route;

    route = New (rip_route_t);
    route->prefix = Ref_Prefix (prefix);
    route->attr = NULL;
    route->time = 0;
    route->delete = 0;
    route->flags = RT_RIP_CHANGE;
    route->imported = rip_ref_attr (attr);

    HASH_Insert (rip->hash, route);
    rip->num_routes++;
    return (route);
}


/* 
 * free route memory and remove from hash
 */
void
rip_delete_route (rip_t * rip, rip_route_t * route)
{
    if (route->imported && route->attr) {
	rip_deref_attr (route->attr);
	route->attr = NULL;
	route->time = 0;
	route->delete = 0;
	route->flags = 0;
    }
    else {
	HASH_Remove (rip->hash, route);
	Deref_Prefix (route->prefix);
	if (route->imported)
	    rip_deref_attr (route->imported);
	if (route->attr)
	    rip_deref_attr (route->attr);
	Delete (route);
	rip->num_routes--;
    }
}


static void
rip_set_flash_update (rip_t * rip)
{
    time_t t;

    pthread_mutex_lock (&rip->flash_update_mutex_lock);
    if (rip->flash_update_waiting) {
	trace (TR_TRACE, rip->trace, "flash timer already running\n");
	pthread_mutex_unlock (&rip->flash_update_mutex_lock);
	return;
    }
/* check to see the RIP and RIPNG */
    t = RIP_FLASH_DELAY_MIN +
	rand () % (RIP_FLASH_DELAY_MAX - RIP_FLASH_DELAY_MIN + 1);
    Timer_Set_Time (rip->flash, t);
    rip->flash_update_waiting++;
    pthread_mutex_unlock (&rip->flash_update_mutex_lock);
}


static void
rip_receive_update (rip_t * rip)
{
#ifdef HAVE_IPV6
    if (rip->proto == PROTO_RIPNG) {
	ripng_receive_update ();
    }
    else
#endif /* HAVE_IPV6 */
    if (rip->proto == PROTO_RIP) {
	rip2_receive_update ();
    }
    else {
	assert (0);
    }
}

static u_int
rip_hash_fn (prefix_t * prefix, u_int size)
{
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6) {
	u_int tmp, buff[4];
	IN6_ADDR_COPY ((struct in6_addr *) buff, prefix_toaddr6 (prefix));
	netmasking (prefix->family, (char *) buff, prefix->bitlen);
	/* Pedro's suggestion */
	tmp = buff[0] ^ buff[1] ^ buff[2] ^ buff[3];
	tmp ^= (tmp >> 16);
	tmp = tmp % size;
	return (tmp);
    }
    else
#endif /* HAVE_IPV6 */
    if (prefix->family == AF_INET) {
	u_int tmp;
	u_char dest[4];
	memcpy (dest, prefix_tochar (prefix), 4);
	netmasking (prefix->family, dest, prefix->bitlen);
	tmp = dest[0] + dest[1] + dest[2] + dest[3];
	tmp = tmp % size;
	return (tmp);
    }
    else {
	assert (0);
    }
    /* NEVER REACHES */
    return (0);
}


static u_int
rip_lookup_fn (prefix_t * tmp1, prefix_t * tmp2)
{
    return (prefix_compare (tmp1, tmp2));
}


static void
rip_update_call_fn (rip_t * rip, int cmd, prefix_t * prefix, rip_attr_t * attr)
{
    if (rip->update_call_fn) {
	generic_attr_t *gattr;
	gattr = New_Generic_Attr (attr->type);
	gattr->gateway = attr->gateway;
	gattr->nexthop = Ref_Prefix (attr->nexthop);
	rip->update_call_fn (cmd, prefix, gattr, attr->pref, 0);
	Deref_Generic_Attr (gattr);
    }
}



/*
 * 1) timoute routes (change metric to 16 and set change bit)
 * 2) garbage collect (delete route from hash, free memory and notify
 *    routing table 
 * Don't need to explicitly do outbound processing, as this will be picked up
 * by the regular 30 second timer
 */
void
rip_timeout_routes (rip_t * rip)
{
    rip_route_t *route;
    LINKED_LIST *ll_delete = NULL;
    time_t now, t;
    time_t nexttime = 0;
    int changed = 0;

    trace (TR_TRACE, rip->trace, "timer (age) fired\n");

    time (&now);
    nexttime = now + RIP_TIMEOUT_INTERVAL;

    pthread_mutex_lock (&rip->route_table_mutex_lock);

    HASH_Iterate (rip->hash, route) {

	if (route->attr == NULL && route->imported &&
	    BIT_TEST (route->flags, RT_RIP_DELETE)) {
	    /* external route being deleted */

	    assert (route->imported->metric == RIP_METRIC_INFINITY);

	    if (now - route->delete <= RIP_GARBAGE_INTERVAL) {
		/* see the earliest next time to be checked */
		goto next;
	    }

	    trace (TR_TRACE, rip->trace,
		   "deleted %s via %s proto %s (garbage collection)\n",
		   prefix_toax (route->prefix),
		   prefix_toa (route->imported->gateway->prefix),
		   proto2string (route->imported->type));
	    if (ll_delete == NULL)
		ll_delete = LL_Create (0);
	    LL_Add (ll_delete, route);
	    continue;
	}

	/* aging RIP route only */
	if (route->attr == NULL)
	    continue;

	/* garbage collect and delete route */
	if (BIT_TEST (route->flags, RT_RIP_DELETE)) {
	    if (now - route->delete >= RIP_GARBAGE_INTERVAL) {
		trace (TR_TRACE, rip->trace,
		       "deleted %s via %s (garbage collection)\n",
		       prefix_toax (route->prefix),
		       prefix_toa (route->attr->gateway->prefix));
		if (ll_delete == NULL)
		    ll_delete = LL_Create (0);
		LL_Add (ll_delete, route);
	    }
	    else {
		goto next;
	    }
	}
	/* timeout route -- set metric to 16 and set change flag */
	else if (now - route->time >= RIP_TIMEOUT_INTERVAL) {
	    trace (TR_TRACE, rip->trace,
		   "timing out %s via %s\n",
		   prefix_toax (route->prefix),
		   prefix_toa (route->attr->gateway->prefix));

	    rip_update_call_fn (rip, RIB_DEL,	/* XXX HOLD */
				route->prefix, route->attr);
	    /* XXX deletion may bring a new imported route,
	       but it will be informed later.
	       So, this may result in sending out the route with infinity */

	    if (route->imported) {
		/* delete ripng route because there is an imported one
		   being announced */
		rip_delete_route (rip, route);
	    }
	    else {
		route->attr->metric = RIP_METRIC_INFINITY;
		route->attr->pref = HOLD_PREF;
		route->flags |= RT_RIP_DELETE;
		route->flags |= RT_RIP_CHANGE;
		route->delete = now;
		changed++;
	    }

	    goto next;
	}
	else {			/* live routes */
	  next:
	    /* see the earliest next time to be checked */
	    if ((t = route->time + RIP_TIMEOUT_INTERVAL) < nexttime)
		nexttime = t;
	}
    }

    /* we have to use a linked list cause we can't delete hash items
     * while iterating through hash table */
    if (ll_delete != NULL) {
	LL_Iterate (ll_delete, route) {
	    rip_delete_route (rip, route);
	}
	LL_Destroy (ll_delete);
    }

    pthread_mutex_unlock (&rip->route_table_mutex_lock);

    if (changed)
	rip_set_flash_update (rip);

    if ((t = nexttime - time (NULL)) <= 0)
	t = 5;			/* don't want so strict? */
    Timer_Set_Time (rip->age, t);
}


/* 
 * announce all change routes to our peers
 */
static void
rip_output_processing (rip_t * rip, int change)
{
    interface_t *interface;
    rip_route_t *route;

    /* on flushing changes, this keeps locking the route table for a long time
       do we need to change our code ? XXX */

    if (change)
	pthread_mutex_lock (&rip->route_table_mutex_lock);

    /* announce routes */
    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	/* check if ripng is configured for this interface */
	if (BIT_TEST (interface->bit, rip->interface_mask) &&
	    BIT_TEST (interface->flags, IFF_UP)) {
#ifdef HAVE_IPV6
	    if (rip->proto == PROTO_RIPNG)
		ripng_send_routes (interface, NULL, change);
	    else
#endif /* HAVE_IPV6 */
		rip2_send_routes (interface, NULL, change);
	}
    }
    /* clearing change flag */
    HASH_Iterate (rip->hash, route) {
	route->flags &= ~(RT_RIP_CHANGE);
    }

    if (change)
	pthread_mutex_unlock (&rip->route_table_mutex_lock);
}


int
check_networks (prefix_t * prefix, LINKED_LIST * ll_networks)
{
    prefix_t *network;

    assert (ll_networks);
    LL_Iterate (ll_networks, network) {
	if (network->family == prefix->family &&
	    a_include_b (network, prefix))
	    return (1);
    }
    return (0);
}


/* 
 * run policy on a rip route 
 * returns -1 when policy rejects it
 * otherwise returns new metric
 */
int
rip_policy (rip_t * rip, prefix_t * prefix, rip_attr_t * attr,
	    interface_t * out)
{
    int cost, num;

#ifdef HAVE_IPV6
    if (rip->proto == PROTO_RIPNG) {
	struct in6_addr *addr = (struct in6_addr *) prefix_tochar (prefix);

	if (prefix->bitlen > 128) {
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (invalid prefix length)\n",
		   prefix_toax (prefix), attr->metric);
	    return (-1);
	}

	/* see if the destination prefix valid */
	if (IN6_IS_ADDR_MULTICAST (addr) ||
	    (IN6_IS_ADDR_V4COMPAT (addr) && prefix->bitlen > 0) ||
	    IN6_IS_ADDR_LINKLOCAL (addr)) {
	    trace (TR_PACKET, RIPNG->trace,
		   "  x %s metric %d (non global)\n",
		   prefix_toax (prefix), attr->metric);
	}
    }
    else
#endif /* HAVE_IPV6 */
    if (rip->proto == PROTO_RIP) {
	if (prefix->bitlen > 32) {
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (invalid prefix length)\n",
		   prefix_toax (prefix), attr->metric);
	    return (-1);
	}
    }
    else {
	assert (0);
    }


    if (out == NULL) {
	/* input policy processing */

	if (attr->metric < 1 || attr->metric > RIP_METRIC_INFINITY) {
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (invalid metric)\n",
		   prefix_toax (prefix), attr->metric);
	    return (-1);
	}

	assert (attr->gateway);
	assert (attr->gateway->interface);
	/* check distribute-list for in */
	if ((num = attr->gateway->interface->dlist_in[rip->proto])) {
	    if (apply_access_list (num, prefix) == 0) {
		trace (TR_PACKET, rip->trace,
		       "  x %s metric %d (a-list %d)\n",
		       prefix_toax (prefix), attr->metric, num);
		return (-1);
	    }
	}
	cost = attr->metric + attr->gateway->interface->metric_in;
    }
    else {
	if (attr->type == PROTO_CONNECTED &&
	    check_networks (prefix, rip->ll_networks)) {
	    /* OK */
	}
	else if (attr->type != rip->proto &&
		 !BIT_TEST (MRT->redist[attr->type], (1 << rip->proto))) {
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (proto %s)\n",
		   prefix_toax (prefix), attr->metric,
		   proto2string (attr->type));
	    return (-1);
	}
	/* split horizon (without poisoned reverse) */
	/* do not send to the same interface on which we got the route */
	if (attr->type == rip->proto &&
	/* multicast capable and p-to-p interfaces */
	/* I'm not sure how NBMA will be supported */
	    ((BIT_TEST (out->flags, IFF_MULTICAST) ||
	      BIT_TEST (out->flags, IFF_POINTOPOINT)) &&
	     attr->gateway->interface == out)) {
	    /* No split horizon on NBMA which has not yet supported, though */
#ifdef RIP_POISONED_REVERSE
	    trace (TR_PACKET, rip->trace,
		   "  o %s metric %d (poisoned reverse)\n",
		   prefix_toax (prefix), RIP_METRIC_INFINITY);
	    return (RIP_METRIC_INFINITY);
#else
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (split horion)\n",
		   prefix_toax (prefix), attr->metric);
	    return (-1);
#endif /* RIP_POISONED_REVERSE */
	}
	if (attr->type == PROTO_CONNECTED && attr->gateway->interface == out) {
	    trace (TR_PACKET, rip->trace,
		   "  x %s metric %d (direct)\n",
		   prefix_toax (prefix), attr->metric);
	    return (-1);
	}
	/* check distribute-list for out */
	if ((num = out->dlist_out[rip->proto])) {
	    if (apply_access_list (num, prefix) == 0) {
		trace (TR_PACKET, rip->trace,
		       "  x %s metric %d (a-list %d)\n",
		       prefix_toax (prefix), attr->metric, num);
		return (-1);
	    }
	}
	cost = attr->metric + out->metric_out;
    }

    if (cost > RIP_METRIC_INFINITY)
	cost = RIP_METRIC_INFINITY;

    return (cost);
}


/* 
 * process linked list of routes (prefix & attribute combinations)
 * check if 1) just updating time on already received route
 *          2) or a change (different attributes from same gateway,
 *             or better metric and attributes from new gateway
 */
int
rip_process_update (rip_t * rip, LINKED_LIST * ll_prefixes,
		    LINKED_LIST * ll_attr)
{
    prefix_t *prefix;
    rip_attr_t *rip_attr;
    rip_route_t *route;
    int metric, change = 0;
    char nexthop[MAXLINE], tag[MAXLINE];
    time_t now;

    time (&now);
    pthread_mutex_lock (&rip->route_table_mutex_lock);

    rip_attr = LL_GetHead (ll_attr);
    LL_Iterate (ll_prefixes, prefix) {
	assert (rip_attr->ref_count > 0);
	assert (prefix->ref_count > 0);

	/* do checks here -- if fail, log and ignore
	 * 1) destination prefix valid (i.e. not multicast)
	 * 2) prefix length valid 
	 * 3) metric valid (between 1 and 16 inclusive)
	 */

	/* see if valid and do policy to add cost to the metric */
	if ((metric = rip_policy (rip, prefix, rip_attr, NULL)) < 0) {
	    continue;
	}
	rip_attr->metric = metric;

#ifdef HAVE_IPV6
#ifdef notdef
	IN6_ADDR_COPY (&in6, prefix_toaddr6 (prefix));
	netmasking (prefix->family, addr, prefix->bitlen);
	if (IN6_ADDR_COMP (&in6, prefix_toaddr6 (prefix)) != 0) {
	    trace (TR_PACKET, rip->trace,
		   "prefix %s will be truncated by prefixlen %d\n",
		   inet_ntop (AF_INET6, &in6, tmp6, sizeof tmp6),
		   prefix->bitlen);
	}
#endif
/*XXX */
	if (prefix_compare2 (rip_attr->nexthop,
			     rip_attr->gateway->prefix) != 0) {
	    sprintf (nexthop, " -> %s",
		     prefix_toa (rip_attr->nexthop));
	}
	else {
	    nexthop[0] = '\0';
	}
#endif /* HAVE_IPV6 */
	if (rip_attr->tag) {
	    sprintf (tag, " tag 0x%lx", rip_attr->tag);
	}
	else {
	    tag[0] = '\0';
	}

	/* new route */
	if ((route = HASH_Lookup (rip->hash, prefix)) == NULL) {

	    if (rip_attr->metric == RIP_METRIC_INFINITY) {
		trace (TR_PACKET, rip->trace,
		       "  x %s metric %d (infinity)\n",
		       prefix_toax (prefix), rip_attr->metric);
		goto next;
	    }

	    trace (TR_TRACE, rip->trace,
		   "  o %s metric %d%s (new)\n",
		   prefix_toax (prefix), rip_attr->metric, tag);

	    route = rip_new_route (rip, prefix, rip_attr);
	    route->flags |= RT_RIP_CHANGE;
	    change++;

	    /* add route to rib */
	    rip_update_call_fn (rip, RIB_UPD, route->prefix,
				route->attr);
	}
	/* exiting route */
	else {

	    if (route->attr == NULL) {
		assert (route->imported);

		if (rip_attr->metric == RIP_METRIC_INFINITY) {
		    trace (TR_PACKET, rip->trace,
			   "  x %s metric %d (infinity)\n",
			   prefix_toax (prefix), rip_attr->metric);
		    goto next;
		}

		route->attr = rip_ref_attr (rip_attr);
		route->time = now;

#ifdef notdef
		if (route->attr->pref <= route->imported->pref ||
		    route->imported->type == PROTO_KERNEL) {
		    trace (TR_PACKET, rip->trace,
			   "  o %s metric %d%s (overwrites %s)\n",
			   prefix_toax (prefix),
			   route->attr->metric, tag,
			   proto2string (route->imported->type));
		    rip_deref_attr (route->imported);
		    route->imported = NULL;
		    route->flags |= RT_RIP_CHANGE;
		    change++;
		}
		else {
#else
		{
#endif
		    if (BIT_TEST (route->flags, RT_RIP_DELETE)) {
			route->flags &= ~RT_RIP_DELETE;
			rip_deref_attr (route->imported);
			route->imported = NULL;
			route->flags |= RT_RIP_CHANGE;
			change++;
			trace (TR_PACKET, rip->trace,
			       "  o %s metric %d%s (%s active)\n",
			       prefix_toax (prefix),
			       route->attr->metric, tag,
			       proto2string (route->attr->type));
		    }
		    else {
			trace (TR_PACKET, rip->trace,
			       "  = %s metric %d%s (%s active)\n",
			       prefix_toax (prefix),
			       route->attr->metric, tag,
			       proto2string (route->attr->type));
		    }
		}

		rip_update_call_fn (rip, RIB_UPD, route->prefix,
				    route->attr);
		goto next;
	    }

	    /* we already have a route for this prefix. Two cases:
	     *  1) from same gateway -- just update time, delete, or implicit withdraw.
	     *     if tag changed it should be updated and propagated -- masaki
	     *  2) from different gateway -- if metric better use this, otherwise
	     *     just ignore it. We'll hear about it again in 30 seconds...
	     */

#if 0
	    if (route->attr->type == PROTO_CONNECTED) {
		trace (TR_PACKET, rip->trace,
		       "  x %s metric %d%s (direct)\n",
		       prefix_toax (prefix),
		       route->attr->metric, tag);
		goto next;
	    }

	    /* shoud check the preferences */
	    if (route->attr->type == PROTO_STATIC) {
		trace (TR_PACKET, rip->trace,
		       "  x %s metric %d%s (static)\n",
		       prefix_toax (prefix),
		       route->attr->metric, tag);
		goto next;
	    }

	    /* kernel route always overriden -- use pref instead */
	    if (route->attr->type == PROTO_KERNEL &&
		rip_attr->metric < RIP_METRIC_INFINITY) {
		trace (TR_PACKET, rip->trace,
		       "  o %s%s metric %d%s (kernel)\n",
		       prefix_toax (prefix),
		       nexthop, route->attr->metric, tag);
		goto overwrite;
	    }
#endif

	    if (route->attr->gateway == rip_attr->gateway) {

		/* case i: both infinity */
		if (rip_attr->metric == RIP_METRIC_INFINITY &&
		    route->attr->metric == RIP_METRIC_INFINITY) {

		    trace (TR_PACKET, rip->trace,
			   "  x %s metric %d (deleting)\n",
			   prefix_toax (prefix), rip_attr->metric);

		    /* it is under deletion process, don't touch. 
		       ignoring it */
		    assert (route->flags & RT_RIP_DELETE);
		    /* should I propagate it? */
		    /* route->flags |= RT_RIP_CHANGE; */
		    /* change++; */
		}
		/* case ii: only new is infinity */
		else if (rip_attr->metric == RIP_METRIC_INFINITY) {

		    trace (TR_TRACE, rip->trace,
			   "  o %s%s metric %d%s (shift to delete)\n",
			   prefix_toax (prefix),
			   nexthop, rip_attr->metric, tag);

		    rip_update_call_fn (rip, RIB_DEL,	/* XXX HOLD */
					route->prefix, route->attr);
		    /* XXX deletion may bring a new imported route,
		       but it will be informed later.
		       So, this may result in sending out the route 
		       with infinity */

		    if (route->imported) {
			rip_delete_route (rip, route);
		    }
		    else {
			/* shift it to deletion process */
			rip_deref_attr (route->attr);
			route->attr = rip_ref_attr (rip_attr);
			route->attr->pref = HOLD_PREF;
			route->flags |= RT_RIP_DELETE;
			route->flags |= RT_RIP_CHANGE;
			route->delete = now;
			change++;
		    }

		}
		/* case iii: new is not infinity (old may be infinity) */
		else {

		    if (route->attr->metric != rip_attr->metric ||
		    /* ID says nothing about tag but I think it should be */
			route->attr->tag != rip_attr->tag ||
			!prefix_compare (route->attr->nexthop,
					 rip_attr->nexthop)) {

			trace (TR_TRACE, rip->trace,
			       "  o %s%s metric %d%s (change)\n",
			       prefix_toax (prefix),
			       nexthop, rip_attr->metric, tag);

			if (!prefix_compare (route->attr->nexthop,
					     rip_attr->nexthop)) {
			    rip_update_call_fn (rip, RIB_UPD,
						route->prefix, rip_attr);
			}
			else {
			    /* the same nexthop but stop deleting process */
			    if (BIT_TEST (route->flags, RT_RIP_DELETE)) {
				/* XXX UNHOLD */
				rip_update_call_fn (rip, RIB_UPD,
						    route->prefix, rip_attr);
			    }
			}

			route->flags &= ~(RT_RIP_DELETE);
			if (route->imported == NULL) {
			    route->flags |= RT_RIP_CHANGE;
			    change++;
			}
			rip_deref_attr (route->attr);
			route->attr = rip_ref_attr (rip_attr);
		    }
		    else {
			trace (TR_PACKET, rip->trace,
			       "  x %s metric %d (same)\n",
			       prefix_toax (prefix),
			       route->attr->metric);
		    }
		    route->time = now;
		}
	    }
	    /* from a gateway different for the existing one */
	    else {

		if (		/* better metric */
		       route->attr->metric > rip_attr->metric ||
		       (route->attr->metric != RIP_METRIC_INFINITY &&
		/* suggested heuristic in case of the same metric */
			(route->attr->metric == rip_attr->metric &&
			 (now - route->time) >=
			 RIP_TIMEOUT_INTERVAL / 2))) {

		    trace (TR_TRACE, rip->trace,
			   "  o %s%s metric %d%s (change)\n",
			   prefix_toax (prefix),
			   nexthop, rip_attr->metric, tag);
		    if (!prefix_compare (route->attr->nexthop,
					 rip_attr->nexthop)) {
#ifdef notdef
			/* add new one then remove old one */
			rip_update_call_fn (rip, RIB_DEL,
					    route->prefix, route->attr);
#endif

			/* don't remove kernel routes. 
			   remove RIP routes only. */
		      overwrite:
			rip_update_call_fn (rip, RIB_UPD,
					    route->prefix, rip_attr);
		    }
		    else {
			/* the same nexthop but metric changed from infinity */
			if (BIT_TEST (route->flags, RT_RIP_DELETE)) {
			    /* XXX UNHOLD */
			    rip_update_call_fn (rip, RIB_UPD,
						route->prefix, rip_attr);
			}
		    }

		    route->time = now;
		    route->flags &= ~(RT_RIP_DELETE);
		    if (route->imported == NULL) {
			route->flags |= RT_RIP_CHANGE;
			change++;
		    }
		    rip_deref_attr (route->attr);
		    route->attr = rip_ref_attr (rip_attr);
		}
		else {
		    trace (TR_PACKET, rip->trace,
			   "  x %s metric %d (>= existing %d)\n",
			   prefix_toax (prefix),
			   rip_attr->metric, route->attr->metric);
		}
	    }
	}
      next:
	rip_attr = LL_GetNext (ll_attr, rip_attr);
    }

    pthread_mutex_unlock (&rip->route_table_mutex_lock);

    if (change)
	rip_set_flash_update (rip);
    return (1);
}



/*
 * rip_update_route()
 * will be called from rib, that is, they are called from
 * other threads. Even though RIPng runs with a single thread,
 * locking on the route table and flash update flag have to be
 * protected by locking.
 *
 * -> changes to use the scheduler. So, I think locking is no need but... 
 */

static void
rip_import (rip_t * rip, prefix_t * prefix, generic_attr_t * new,
	    generic_attr_t * old, int pref)
{
    rip_attr_t *rip_attr;
    rip_route_t *route;
    time_t now;

    assert (new == NULL || new->type != rip->proto);
    assert (old == NULL || old->type != rip->proto);

    pthread_mutex_lock (&rip->route_table_mutex_lock);
    time (&now);

    if (old) {

	trace (TR_TRACE, rip->trace,
	       "delete %s nh %s proto %s\n",
	       prefix_toax (prefix),
	       prefix_toa (old->nexthop),
	       proto2string (old->type));

	if ((route = HASH_Lookup (rip->hash, prefix)) == NULL) {
	    trace (TR_TRACE, rip->trace, "no entry in the table!\n");
	    goto quit;
	}

	if (route->imported == NULL) {
	    trace (TR_TRACE, rip->trace, "no imported entry!\n");
	    goto quit;
	}

	if (route->imported->type != old->type)
	    trace (TR_TRACE, rip->trace,
		   "entry was learned from proto %s\n",
		   proto2string (route->imported->type));

	if (route->attr) {
	    rip_deref_attr (route->imported);
	    route->imported = NULL;
	}
	else {
	    route->flags |= RT_RIP_DELETE;
	    route->imported->metric = RIP_METRIC_INFINITY;
	    route->flags |= RT_RIP_CHANGE;
	    route->delete = now;

	    rip_set_flash_update (rip);
	}
    }

    if (new) {

	rip_attr = rip_new_attr (rip, 1 /* XXX */ );
	rip_attr->type = new->type;
	rip_attr->nexthop = Ref_Prefix (new->nexthop);
	rip_attr->gateway = new->gateway;
	rip_attr->pref = pref;
	rip_attr->tag = new->tag;

	if ((route = HASH_Lookup (rip->hash, prefix)) == NULL) {
	    /* new route */
	    trace (TR_TRACE, rip->trace,
		   "add %s nh %s proto %s (new)\n",
		   prefix_toax (prefix),
		   prefix_toa (new->nexthop),
		   proto2string (new->type));
	    route = rip_new_imported_route (rip, prefix, rip_attr);
	}
	else {
	    if (route->attr && route->attr->pref < pref) {
		trace (TR_TRACE, rip->trace,
		       "active better? %s nh %s proto %s (pref %d)\n",
		       prefix_toax (prefix),
		       prefix_toa (new->nexthop),
		       proto2string (route->attr->type), route->attr->pref);
		/* goto quit; */
	    }
	    trace (TR_TRACE, rip->trace,
		   "add %s nh %s proto %s (overwrite %s)\n",
		   prefix_toax (prefix),
		   prefix_toa (new->nexthop),
		   proto2string (new->type),
		   proto2string ((route->imported) ? route->imported->type :
				 route->attr->type));
	    if (route->imported)
		rip_deref_attr (route->imported);
	    route->imported = rip_ref_attr (rip_attr);
	    if (route->attr == NULL)
		route->flags &= ~RT_RIP_DELETE;
	}

	route->flags |= RT_RIP_CHANGE;
	rip_deref_attr (rip_attr);

	if (new->type != PROTO_STATIC &&
	    new->type != PROTO_CONNECTED &&
	    new->type != PROTO_KERNEL) {
	    rip_set_flash_update (rip);
	}
    }

  quit:
    Deref_Prefix (prefix);
    if (new)
	Deref_Generic_Attr (new);
    if (old)
	Deref_Generic_Attr (old);
    pthread_mutex_unlock (&rip->route_table_mutex_lock);
}


int
rip_update_route (prefix_t * prefix, generic_attr_t * new,
		  generic_attr_t * old, int pref)
{
    rip_t *rip = RIP;

    /* in case the rib wants to delete them */
    Ref_Prefix (prefix);
    if (new)
	Ref_Generic_Attr (new);
    if (old)
	Ref_Generic_Attr (old);

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    schedule_event2 ("rip_import",
		 rip->schedule, rip_import, 5, rip, prefix, new, old, pref);
    return (1);
}


/* 
 * change/set rip attributes.
 */
void
rip_set (int proto, int first,...)
{
    va_list ap;
    enum RIP_ATTR attr;
    rip_t *rip = RIP;

#ifdef HAVE_IPV6
    if (proto == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    /* Process the Arguments */
    va_start (ap, first);

    for (attr = (enum RIP_ATTR) first;
	 attr;
	 attr = va_arg (ap, enum RIP_ATTR)) {

	switch (attr) {
	case RIP_RT_UPDATE_FN:
	    rip->update_call_fn = va_arg (ap, int_fn_t);
	    break;
	case RIP_TRACE_STRUCT:
	    rip->trace = va_arg (ap, trace_t *);
	    break;
	default:
	    assert (0);
	    break;
	}
    }
    va_end (ap);
}


/* 
 * dump various rip stats to a socket
 * usually called by UII (user interactive interface 
 */
int
rip_show (int proto, uii_connection_t * uii)
{
    rip_t *rip = RIP;
    int port = RIP_DEFAULT_PORT;

#ifdef HAVE_IPV6
    if (proto == PROTO_RIPNG) {
	rip = RIPNG;
	port = RIPNG_DEFAULT_PORT;
    }
#endif /* HAVE_IPV6 */
    uii_send_data (uii, "\r\nRouting Protocol is \"%s\"\r\n",
		   proto2string (rip->proto));

    if (rip->sockfd < 0)
	uii_send_data (uii, "Not listening for announcements\r\n");
    else
	uii_send_data (uii, "Listening on port %d (socket %d)\r\n",
		       port, rip->sockfd);

    uii_send_data (uii,
      "Sending updates every %d seconds +/- %d, next due in %d seconds\r\n",
		   RIP_UPDATE_INTERVAL, RIP_UPDATE_JITTER,
		   rip->timer->time_next_fire - time (NULL));
    uii_send_data (uii, "Triggered update and split horizon "
		   "(no poisoned reverse) implemented\r\n");
    uii_send_data (uii,
	     "Invalid after %d seconds, hold down %d, flushed after %d\r\n",
	  RIP_TIMEOUT_INTERVAL, RIP_TIMEOUT_INTERVAL, RIP_GARBAGE_INTERVAL);

    uii_send_data (uii, "\r\n");
    uii_send_data (uii, "%d routes ", rip->num_routes);
    uii_send_data (uii, "%d hash entries\r\n", HASH_GetCount (rip->hash));

    return (1);
}


static int
rip_route_compare (rip_route_t * a, rip_route_t * b)
{
    return (prefix_compare2 (a->prefix, b->prefix));
}


/*
 * dump routing table to socket. Usually called by user interactive interface
 */
int
rip_show_routing_table (int proto, uii_connection_t * uii)
{
    rip_route_t *route;
    time_t now, t;
    interface_t *interface;
    char tmp[MAXLINE];
    char *name = NULL, save[MAXLINE];
    DATA_PTR *array;
    int i, c;
    u_int nel;
    rip_t *rip = RIP;

#ifdef HAVE_IPV6
    if (proto == PROTO_RIPNG) {
	rip = RIPNG;
    }
#endif /* HAVE_IPV6 */

    if (parse_line (uii->cp, "%s", save) >= 1) {
	if ((interface = find_interface_byname (save)) == NULL) {
	    uii_send_data (uii, "no such interface: %s\r\n", save);
	    return (-1);
	}
	name = save;
    }

    pthread_mutex_lock (&rip->route_table_mutex_lock);
    time (&now);
    array = HASH_ToArray (rip->hash, NULL, &nel);
    pthread_mutex_unlock (&rip->route_table_mutex_lock);
    ARRAY_Sort (array, nel, rip_route_compare);

    sprintf (tmp, "%-4s %-4s", "Cost", "Time");
    rib_show_route_head (uii, tmp);

    for (i = 0; i < nel; i++) {

	route = (rip_route_t *) array[i];

	if (route->attr) {
	    rip_attr_t *attr = route->attr;

	    if (name == NULL || (attr->gateway->interface &&
		      strcmp (attr->gateway->interface->name, name) == 0)) {

		c = (route->imported &&
		     route->imported->pref < route->attr->pref) ?
		    ' ' : '*',
		    t = now - attr->time;
		sprintf (tmp, "%4d %4ld", attr->metric,
			 now - route->time);

		rib_show_route_line (uii, c, ' ', attr->type,
				     attr->pref, t,
				     route->prefix, attr->nexthop,
				     attr->gateway, tmp);
	    }
	}

	if (route->imported) {
	    rip_attr_t *attr = route->imported;
	    t = now - route->imported->time;

	    if (name == NULL || (attr->gateway->interface &&
		      strcmp (attr->gateway->interface->name, name) == 0)) {

		c = (route->attr &&
		     route->attr->pref < route->imported->pref) ?
		    '?' : '*';
		if (!check_networks (route->prefix, rip->ll_networks) &&
		    !BIT_TEST (MRT->redist[attr->type], 1 << rip->proto))
		    c = 'x';

		t = now - attr->time;
		sprintf (tmp, "%4d ----", attr->metric);

		rib_show_route_line (uii, c, ' ', attr->type,
				     attr->pref, t,
				     route->prefix, attr->nexthop,
				     attr->gateway, tmp);
	    }
	}
    }
    Delete (array);
    return (1);
}


static void
rip_timer_update (rip_t * rip)
{
    trace (TR_TRACE, rip->trace, "timer (update) fired\n");
    pthread_mutex_lock (&rip->flash_update_mutex_lock);
    if (rip->flash_update_waiting)
	rip->flash_update_waiting = 0;	/* clear flash update */
    pthread_mutex_unlock (&rip->flash_update_mutex_lock);
    rip_output_processing (rip, 0);
}


static void
rip_flash_update (rip_t * rip)
{
    trace (TR_TRACE, rip->trace, "timer (flash update) fired\n");
    pthread_mutex_lock (&rip->flash_update_mutex_lock);
    if (rip->flash_update_waiting) {
	rip->flash_update_waiting = 0;
	pthread_mutex_unlock (&rip->flash_update_mutex_lock);
	rip_output_processing (rip, 1);
    }
    else {
	pthread_mutex_unlock (&rip->flash_update_mutex_lock);
    }
}


/*
 * called from select on socket
 */
void
rip_schedule_receive_update (rip_t * rip)
{
    schedule_event2 ("rip_receive_update",
		     rip->schedule, rip_receive_update, 1, rip);
}


/*
 * called from timer
 */

/* 30 timer processing -- announce routing table */
static void
rip_schedule_timer (mtimer_t * timer, rip_t * rip)
{
    schedule_event2 ("rip_timer_update",
		     rip->schedule, rip_timer_update, 1, rip);
}


/* holddown & expire */
static void
rip_schedule_age (mtimer_t * timer, rip_t * rip)
{
    schedule_event2 ("rip_timeout_routes",
		     rip->schedule, rip_timeout_routes, 1, rip);
}


/* flash update */
static void
rip_schedule_flash (mtimer_t * timer, rip_t * rip)
{
    schedule_event2 ("rip_flash_update",
		     rip->schedule, rip_flash_update, 1, rip);
}


/*
 * calling sequence: init_rip and then start_rip
 */

/*
 * initialize rip stuff
 */
int
rip_init (int proto, trace_t * tr)
{
    rip_route_t tmp;
    rip_t *rip;

    rip = New (rip_t);
    rip->trace = trace_copy (tr);
    rip->proto = proto;

#ifdef HAVE_IPV6
    if (proto == PROTO_RIPNG) {
	set_trace (rip->trace, TRACE_PREPEND_STRING, "RIPNG", 0);
	rip->schedule = New_Schedule ("RIPNG", rip->trace);
	rip->timer = New_Timer (rip_schedule_timer, RIP_UPDATE_INTERVAL,
				"RIPNG update timer", rip);
	rip->age = New_Timer (rip_schedule_age, 0, "RIPNG aging timer", rip);
	rip->flash = New_Timer (rip_schedule_flash, 0, "RIPNG flash timer",
				rip);
	ripng_init ();
	RIPNG = rip;
    }
    else
#endif /* HAVE_IPV6 */
    if (proto == PROTO_RIP) {
	set_trace (rip->trace, TRACE_PREPEND_STRING, "RIP", 0);
	rip->schedule = New_Schedule ("RIP", rip->trace);
	rip->timer = New_Timer (rip_schedule_timer, RIP_UPDATE_INTERVAL,
				"RIP update timer", rip);
	rip->age = New_Timer (rip_schedule_age, 0, "RIP aging timer", rip);
	rip->flash = New_Timer (rip_schedule_flash, 0, "RIP flash timer", rip);
	RIP = rip;
    }
    else {
	assert (0);
    }

    timer_set_jitter (rip->timer, RIP_UPDATE_JITTER);
    timer_set_flags (rip->age, TIMER_ONE_SHOT);
    timer_set_flags (rip->flash, TIMER_ONE_SHOT);

    rip->hash = HASH_Create (100,
			     HASH_KeyOffset, HASH_Offset (&tmp, &tmp.prefix),
			     HASH_LookupFunction, rip_lookup_fn,
			     HASH_HashFunction, rip_hash_fn,
			     NULL);

    pthread_mutex_init (&rip->send_mutex_lock, NULL);
    pthread_mutex_init (&rip->route_table_mutex_lock, NULL);
    pthread_mutex_init (&rip->flash_update_mutex_lock, NULL);
    pthread_mutex_init (&rip->mutex_lock, NULL);

    rip->ll_networks = LL_Create (0);
    rip->ll_interfaces = LL_Create (0);
    rip->ll_dlists = LL_Create (0);

    return (1);
}


static int
rip_init_listen (rip_t * rip)
{
#ifdef HAVE_IPV6
    if (rip->proto == PROTO_RIPNG)
	return (ripng_init_listen ());
    else
#endif /* HAVE_IPV6 */
    if (rip->proto == PROTO_RIP)
	return (rip2_init_listen ());
    else {
	assert (0);
	return (0);
    }
}


static void
rip_send_request (rip_t * rip, LINKED_LIST * ll_prefixes)
{
#ifdef HAVE_IPV6
    if (rip->proto == PROTO_RIPNG)
	ripng_send_request (ll_prefixes);
    else
#endif /* HAVE_IPV6 */
    if (rip->proto == PROTO_RIP)
	rip2_send_request (ll_prefixes);
    else
	assert (0);
}


/* 
 * the thread is alive! Start the timers and wait for events to process
 */
static void
rip_start_main (rip_t * rip)
{
    init_mrt_thread_signals ();
    trace (TR_TRACE, rip->trace, "thread created\n");

    if (rip_init_listen (rip) < 0) {
	trace (TR_ERROR, rip->trace, "aborted due to error(s)\n");
	return;
    }
    Timer_Turn_ON (rip->timer);
    Timer_Set_Time (rip->age, RIP_TIMEOUT_INTERVAL);

    /* send initial request to routers */
    rip_send_request (rip, NULL);

#ifndef HAVE_LIBPTHREAD
    return;
#else
    while (1) {
	schedule_wait_for_event (rip->schedule);
    }
#endif /* HAVE_LIBPTHREAD */
}


/* 
 * just call thread create routine
 */
void
rip_start (int proto)
{
#ifdef HAVE_IPV6
    if (proto == PROTO_RIPNG) {
	mrt_thread_create ("RIPNG thread", RIPNG->schedule,
			   (thread_fn_t) rip_start_main, RIPNG);
    }
    else
#endif /* HAVE_IPV6 */
    if (proto == PROTO_RIP) {
	mrt_thread_create ("RIP thread", RIP->schedule,
			   (thread_fn_t) rip_start_main, RIP);
    }
    else {
	assert (0);
    }
}
